#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/errno.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/io.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/device.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <linux/timer.h>

#include "rpi_gpio.h"

MODULE_AUTHOR("Satoshi Yoneda");
MODULE_LICENSE("Dual BSD/GPL");


#define SSEG_GPIO_MAPNAME	"timertest_gpio_map"

static void __iomem *gpio_map;
static volatile uint32_t *gpio_base;

static struct timer_list refresh_timer;

static int display_refresh_hz = 62;
static int display_digits = 4;

#define LED_BASE	7
#define DIGIT_BASE	14
#define DIGITS		4
#define BLANK		10

static int display_value[DIGITS] = {4,3,2,1};

/* proto types */
static int rpi_gpio_map(void);
static int rpi_gpio_unmap(void);
static int rpi_gpio_function_set(int, uint32_t);
static void rpi_gpio_set32( uint32_t, uint32_t );
static void rpi_gpio_clear32( uint32_t, uint32_t );
static void gpio_setup(void);
static void led_put( unsigned int);
static void register_refresh_timer(void);
static void refresh_timer_handler(unsigned long data);
static int timertest_init(void);
static void timertest_exit(void);


static int rpi_gpio_map(void)
{
	if( !request_mem_region(RPI_GPIO_BASE,
							RPI_BLOCK_SIZE,
							SSEG_GPIO_MAPNAME) ) {
		printk( KERN_ALERT "request_mem_region failed.\n");
		return -EBUSY;
	}
	gpio_map = ioremap_nocache(RPI_GPIO_BASE, BLOCK_SIZE);
	gpio_base = (volatile uint32_t *)gpio_map;
	
	return 0;
}

static int rpi_gpio_unmap(void)
{
	iounmap(gpio_map);
	release_mem_region(RPI_GPIO_BASE, RPI_BLOCK_SIZE);

	gpio_map = NULL;
	gpio_base = NULL;
	return 0;
}

static int rpi_gpio_function_set(int pin, uint32_t func)
{
	int index = RPI_GPFSEL0_INDEX + pin / 10;
	uint32_t mask = ~(0x7 << ((pin % 10) * 3));
	gpio_base[index] = (gpio_base[index] & mask) | ((func & 0x7) << ((pin % 10) * 3));
	
	return 1;
}

static void rpi_gpio_set32( uint32_t mask, uint32_t val )
{
	gpio_base[RPI_GPSET0_INDEX] = val & mask;
}

static void rpi_gpio_clear32( uint32_t mask, uint32_t val )
{
	gpio_base[RPI_GPCLR0_INDEX] = val & mask;
}

static void gpio_setup(void)
{
	rpi_gpio_function_set( LED_BASE+0, RPI_GPF_OUTPUT );
	rpi_gpio_function_set( LED_BASE+1, RPI_GPF_OUTPUT );
	rpi_gpio_function_set( LED_BASE+2, RPI_GPF_OUTPUT );
	rpi_gpio_function_set( LED_BASE+3, RPI_GPF_OUTPUT );

	rpi_gpio_function_set( DIGIT_BASE+0, RPI_GPF_OUTPUT );
	rpi_gpio_function_set( DIGIT_BASE+1, RPI_GPF_OUTPUT );
}

static void led_put( unsigned int v )
{
	rpi_gpio_clear32( RPI_GPIO_P1MASK, 0x0F << LED_BASE);
	rpi_gpio_set32( RPI_GPIO_P1MASK, (v & 0x0F) << LED_BASE);	
}


static void register_refresh_timer(void)
{
	init_timer(&refresh_timer);
	
	refresh_timer.expires = jiffies + HZ / (display_refresh_hz * DIGITS);
	/* 最短のTickより短い場合は最短のTickに */
	if( refresh_timer.expires == 0 ) refresh_timer.expires = jiffies + 1;
	refresh_timer.data = jiffies;
	refresh_timer.function = refresh_timer_handler;
	
	add_timer(&refresh_timer);
}

static void refresh_timer_handler(unsigned long data)
{
	static int dig = 0;
	
	if( ++dig >= DIGITS ) dig = 0;
	
	led_put(BLANK);
	rpi_gpio_clear32( RPI_GPIO_P1MASK, 0x03 << DIGIT_BASE);
	rpi_gpio_set32( RPI_GPIO_P1MASK, (dig) << DIGIT_BASE);
	led_put(display_value[dig]);
	
	register_refresh_timer();
}

static int timertest_init(void)
{
	int retval;
	
	/* 開始のメッセージ */
	printk(KERN_INFO "%s loading...\n", "timertest" );
	
	/* GPIOレジスタがマップ可能か調べる */
	retval = rpi_gpio_map();
	if( retval != 0 ) {
		printk( KERN_ALERT "Can not use GPIO registers.\n");
		return -EBUSY;
	}
	/* GPIO初期化 */
	gpio_setup();
	printk( KERN_INFO "RPI HZ = %d\n", HZ );
	register_refresh_timer();
	printk( KERN_INFO "timertest driver register sccessed.\n");
	
	return 0;
}

static void timertest_exit(void)
{
	del_timer_sync(&refresh_timer);
	rpi_gpio_unmap();
}

module_init(timertest_init);
module_exit(timertest_exit);
